#include <Ticker.h>
#include "mdEspRestartSF.h"

extern "C" {
#include "user_interface.h"
}

/*
 * Returns true if the code being executed has
 * just been downloaded over the UART
 */
bool isUartBoot(void) {
  int bootmode = 0;
  asm (
    "movi %0, 0x60000200\n\t"
    "l32i %0, %0, 0x118\n\t"
    : "+r" (bootmode) /* Output */
    : /* Inputs (none) */                
    : "memory" /* Clobbered */           
  );
  return (((bootmode >> 0x10) & 0x7) == 1); 
}


/*
 * Structure in which restart information will be saved
 * using RTC memory. The latter is divided into 4 byte
 * blocks, the first 64 blocks (256 bytes) are used by the 
 * system, the last 128 (512 bytes) are available to the
 * user. 
 * 
 * Two attributes are defined to control the allocation of
 * memory and alignment of the global structure.
 * 
 * First the members of the structure are ordered to ensure
 * that it occupies exactly one block. The packed attribute 
 * is not needed, but put reason after count and there will 
 * padding will be added to align data on a 4 byte boundary.
 * 
 * The aligned attribute is needed to ensure that the whole 
 * structure is aligned on a 4 byte boundary because  
 * system_rtc_mem_read and system_rtc_write will not work 
 * otherwise. The attribute is needed because the structure
 * will be aligned on 2 byte boundary only as that is the
 * size of its biggest member.
 * 
 * GCC online documentation on attributes:
 * 
 * Specifying Attributes of Variables  
 * https://gcc.gnu.org/onlinedocs/gcc-3.3/gcc/Variable-Attributes.html#Variable%20Attributes
 * 
 * Espressif, ESP8266 Non-OS SDK API Reference, Version 2.2.1 (2018/05)
 * "3.3.23 system_rtc_mem_write" p. 19
 * https://www.espressif.com/sites/default/files/documentation/2c-esp8266_non_os_sdk_api_reference_en.pdf#page=1&zoom=auto,-13,792
 *
 */

struct __attribute__((packed)) {
  restartReason_t reason;   // 1 byte: 4 bit marker, 4 bit restart reason  
  restartCount_t   count;   // 1 byte (8 bit) count
  restartData_t     data;   // 2 byte (16 bit) data (module id or exception cause or user defined)
} restart __attribute__((aligned(4)));

#define RESTART_MARKER    0b10110000 // 0xB0 = 176
#define LWD_USR_MARKER    0b10010000 // 0x90 = 144
#define VALID_MARKER      0b10010000 // 0x90 = 144
#define VALID_MARKER_MASK 0b11010000 // 0xE0 = 224
#define MARKER_MASK       0b11110000 // 0xF0 = 240
#define REASON_MASK       0b00001111 // 0x0F = 15


/*
 * By default the restart structure will be stored at the
 * very end of the RTC memory
 * 
 * The 127 blocks of user RTC memory follow 64 4-byte blocks 
 * (256 bytes) reserved by the system
 */
#define RESTART_BLOCK_SIZE ((sizeof(restart) + 3) & (~3))
#define LAST_VALID_BLOCK_ADDRESS 64 + 128 - (RESTART_BLOCK_SIZE / 4)

uint8 rtcAddress = LAST_VALID_BLOCK_ADDRESS;

bool setRestartRtcAddress(uint8_t addr) {
  bool result = true;
  if (addr == DEFAULT_RESTART_ADDRESS) {
    rtcAddress = LAST_VALID_BLOCK_ADDRESS;
  } else { 
    addr = addr + 64;
    result = (addr <= LAST_VALID_BLOCK_ADDRESS);
    if (result) rtcAddress = addr;
  } 
  return result;
}   

void initRestart(void) {
  restart.reason = RESTART_MARKER;
  restart.count = 0;
  restart.data = LOOP_END;
  //dbg dumpRestart("initRestart()");
}


void loadRestart(void) {
  system_rtc_mem_read(rtcAddress, &restart, sizeof(restart));
  //dbg dumpRestart("loadRestart()");
  if ( ( (restart.reason & VALID_MARKER_MASK) != VALID_MARKER) || ( (restart.reason & REASON_MASK) > LAST_VALID_REASON) ) {
    initRestart();
  }
  
}    


void saveRestart(void)  {
  system_rtc_mem_write(rtcAddress, &restart, sizeof(restart));
  //dbg dumpRestart("saveRestart()");
}


void setLwdtRestartReason(restartReason_t reason) {
  //dbg dumpRestart("setLwdtRestartReason()");
  if ( (restart.reason & REASON_MASK) != reason) {
    restart.count = 0;  // it will be incremented by getRestartReason
  }  
 restart.reason = ( LWD_USR_MARKER | reason );
 saveRestart();
}


/*
 * Flag to avoid reading restart struct from memory more than once,
 * in other words, getRestartReason can be called more than once, on  
 * second or later calls, the result will be the same as the first
 * call.
 */
boolean restartUpdated = false;   // 

restartReason_t getRestartReason(restartCount_t &count, restartData_t &data) {
  if (!restartUpdated) {
    //dbg Serial.println("getRestartReason()");
    loadRestart();
    
    restartReason_t reason = ESP.getResetInfoPtr()->reason;    
    restartReason_t restartreason = (restart.reason & REASON_MASK);

    //dbg Serial.printf("getRestartReason():: reason: %d, restartreason: %d\n", reason, restartreason);
  
    if ( ((restart.reason & MARKER_MASK) == LWD_USR_MARKER ) && (reason == REASON_SOFT_RESTART) )  {
        reason = restartreason;
    }  

    boolean isSameReason = (reason == restartreason);
    if (reason == REASON_EXCEPTION_RST) {
      restartData_t exceptionCause = ESP.getResetInfoPtr()->exccause;
      isSameReason = isSameReason && (exceptionCause == restart.data);
      restart.data = exceptionCause;
    }

    //dbg Serial.printf("getRestartReason():: isSameReason: %s\n", (isSameReason) ? "true" : "false");
    
    if (isSameReason) {
      restart.count++;
    } else {
      restart.count = 1;
    }    
    restart.reason = (RESTART_MARKER | reason);
    saveRestart();
    restartUpdated = true;  
  }
  data = restart.data;
  count = restart.count;
  return (restart.reason & REASON_MASK);
}


void userRestart(restartData_t data) {
  restart.data = data;
  setLwdtRestartReason(REASON_USER_RESTART);
  ESP.restart();
}


void userReset(restartData_t data) {
  restart.data = data;
  setLwdtRestartReason(REASON_USER_RESET);
  ESP.reset();
}

#ifdef INCLUDE_GETRESETREASONEX
String getResetReasonEx(void) {
  char buff[80];
  restartCount_t count;
  restartData_t data;
  restartReason_t restartReason = getRestartReason(count, data);
  if (restartReason == REASON_EXCEPTION_RST) {
    snprintf(buff, sizeof(buff), "%s %d, count %d", ESP.getResetReason().c_str(), data, count);
  } else if (restartReason <= REASON_EXT_SYS_RST) {
    snprintf(buff, sizeof(buff), "%s, count %d", ESP.getResetReason().c_str(), count);
  } else if (restartReason == REASON_USER_RESET) {
    snprintf_P(buff, sizeof(buff), PSTR("User reset, data %d, count %d"), data, count);
  } else if (restartReason == REASON_USER_RESTART) {
    snprintf_P(buff, sizeof(buff), PSTR("User restart, data %d, count %d"), data, count);
  } else if (restartReason == REASON_LWD_RST) {
    snprintf_P(buff, sizeof(buff), PSTR("Loop watchdog in %d, count %d"), data, count);
  } else if (restartReason == REASON_LWD_LOOP_RST) {
    snprintf_P(buff, sizeof(buff), PSTR("Loop watchdog incomplete in %d, count %d"), data, count);
  } else if (restartReason == REASON_LWD_OVW_RST) {
    snprintf_P(buff, sizeof(buff), PSTR("Loop watchdog overwritten in %d, count %d"), data, count);
  } else {
    return ESP.getResetReason();  
  }
  return String(buff);  
}
#endif

Ticker lwdTicker;
volatile unsigned long lwdTime;
volatile unsigned long lwdTimeout;
volatile unsigned long lwdDifference;
volatile unsigned long lwdNotDifference;

/*
 * N O T E
 * 
 * clearRestart() should be invoked when user chooses to restart the
 * system manually ????
 * 
void clearRestart(void) {
 initRestart();
 saveRestart();
}
 */
 
void lwdtStamp(restartData_t data) {
  restart.data = (restartData_t) data;
}
  
void lwdtFeed(void) {
  lwdTime = millis();
  lwdTimeout = lwdTime + lwdDifference;
  if (restart.data != LOOP_END) {
    setLwdtRestartReason(REASON_LWD_LOOP_RST);
    ESP.restart(); 
  }
}

void ICACHE_RAM_ATTR lwdtcb(void) {
  if ( ((lwdTimeout - lwdTime) != lwdDifference) || (lwdDifference != - lwdNotDifference) )
    setLwdtRestartReason(REASON_LWD_OVW_RST);
  else if ( (restart.reason & VALID_MARKER_MASK) != VALID_MARKER) 
    setLwdtRestartReason(REASON_LWD_OVW_RST);           // where variable mangled    
  else if (millis() - lwdTime < LWD_TIMEOUT)
    return;
  else
    setLwdtRestartReason(REASON_LWD_RST);  
  ESP.restart();   // lwd timed out or lwd overwritten
}

void lwdtInit(unsigned long timeout) {
  lwdNotDifference = -timeout;
  lwdDifference = timeout;
  lwdTicker.attach_ms(timeout / 2, lwdtcb); 
  lwdtStamp(LOOP_END); //= start of loop()
  lwdtFeed();
}

#ifdef DEBUG_RTC_PART
void dumpInfo(void) {
  Serial.println("mdEspRestart info");
  Serial.printf("  sizeof(restart): %d\n", sizeof(restart));
  Serial.printf("  RESTART_BLOCK_SIZE: %d\n", RESTART_BLOCK_SIZE);
  Serial.printf("  sizeof(restart) is a multiple of 4 bytes: %s\n", (sizeof(restart)==RESTART_BLOCK_SIZE) ? "true" : "*** FALSE ***");
  Serial.printf("  rtcAddress: %d\n",  rtcAddress);
  Serial.printf("  (rtcAddress-64)*4 + sizeof(restart): %d\n", (rtcAddress-64)*4 + sizeof(restart));
  Serial.printf("  (rtcAddress-64)*4 + RESTART_BLOCK_SIZE: %d\n", (rtcAddress-64)*4 + RESTART_BLOCK_SIZE);

  Serial.printf("  memory location of restart: %p\n", &restart);
  Serial.printf("  restart is 32 bit aligned: %s\n", (((uint32_t) &restart & (uint32_t) 0b11) == 0) ? "true" : "*** FALSE ***");
  boolean ok = system_rtc_mem_read(rtcAddress, &restart, sizeof(restart));
  Serial.printf("  can do system_rtc_mem_read %s\n", (ok) ? "yes" : "**** NO ***");   
  Serial.println();
}

void dumpRestart(char * msg){
  Serial.printf("%s:: reason (with flag): 0x%0x, reason: %d, count: %d, data: 0x%0x\n", msg, restart.reason, (restart.reason & REASON_MASK), restart.count, restart.data);
  Serial.printf("%s:: flag: 0x%0x = ", msg, (restart.reason >> 4));
  if ((restart.reason & MARKER_MASK) == RESTART_MARKER)
    Serial.print("RESTART_MARKER");
  else if ((restart.reason & MARKER_MASK) == LWD_USR_MARKER)
    Serial.print("LWD_USR_MARKER");
  else
    Serial.print("unknown");
  Serial.printf(", valid: %s\n", ((restart.reason & VALID_MARKER_MASK) == VALID_MARKER) ? "true" : "false");
}
#endif

#ifdef DEBUG_LWD_PART
void mangleLwdt(void) {
  lwdTime = 0;
  lwdTimeout = 0;
  lwdDifference = 0;
}
void mangleLwdtFlag(void) {
  restart.reason = (restart.reason | 0b11100000);
}
#endif
